# Attendance ID Checker
# Copyright 2024-01-22 Osamu NAKAMURA

# RUN
#    e.g.1> python thisprogram.py file1.csv file2.csv
#    e.g.2> python thisprogram.py file1.csv file2.csv file3.csv
#    real.> python 03_id_checker_20240123.py id_001.csv moodle_quiz.csv order.csv
#    test.> python 03_id_checker_20240123.py test-04_src_EA2023Class10.csv test-04_moodle_quiz_result.csv
#    test.> python 03_id_checker_20240123.py test_IDs_setA/id_001.csv test_IDs_setA/moodle_c01_answer.csv
#    test.> python 03_id_checker_20240123.py test_IDs_setA/id_001.csv test_IDs_setA/moodle_c01_answer.csv test_IDs_setA/moodle_order.csv
#    test.> python 03_id_checker_20240123.py test_IDs_setA/id_002.csv test_IDs_setA/moodle_c02_answer.csv
#    test.> python 03_id_checker_20240123.py test_IDs_setA/id_002.csv test_IDs_setA/moodle_c02_answer.csv test_IDs_setA/moodle_order.csv

#    Run this program with the following three arguments:
#
# 1) csv file containing the attendance IDs that have already been created.
#      File format:
#      - 1st column: attendance ID  (Other columns ignored)
#      - no header.
#
# 2) csv file of attendance IDs entered by students in Moodle quiz.
#      Downloaded from Moodle quiz.
#      When downloading from Moodle, the following two options must be selected:
#      [1] "Exam Status: Completed"
#      [2] "Show at most one completed attempt per user".
#      Flie format:
#      - Dpending on Moodle.
#      - Columns specified by c_sn1/c_sid/c_aid below.
#      - 1st row: header.
#
# 3) [optional] A csv file with the student IDs to be output.
#      The program will sort the results in the order of the student IDs in
#      this file. If this argument is omitted, the output is in the order
#      of the students in the second argument.
#      File format:
#      - 1st column: student ID     (Other columns ignored)
#      - no header.

# OUTPUT
#    ~_at.csv


import os
import sys
import re
import pandas

c_sn1     =  0   # Column Number of Student Name  in csv output by Moodle
c_sid     =  2   # Column Number of Student ID    in csv output by Moodle
c_aid     = 13   # Column Number of Attendance ID in csv output by Moodle
        # (Count from 0; 0=A, 1=B, 2=C, .. in Excel)

v_arg     = sys.argv
        # read arguments.

f_in1     = v_arg[1]
f_in2     = v_arg[2]
if ( len(   v_arg ) >= 4 ):
    f_in3 = v_arg[3]
else:
    f_in3 = ""


############
############ Data Loading 1
############ Attendance IDs handed out to students
############

dt1    = pandas.read_csv( f_in1, header=None )

#for id in dt1[0]:
#    print(id)

l_ais  = dt1[0].values.tolist()
        # Convert the 1st column of the csv-file to a list.
        # ais: Attendance ID Source.


############
############ Data loading 2
############ Attendance IDs registered by students in Moodle
############

tmp    = pandas.read_csv( f_in2, header=None )
dt2    = tmp[1:]
        # Data is imported as a data frame using pandas.
        # tmp[1:] to drop the 1st line (=header)

        # Note.
        # The csv output by Moodle actually has a header.
        # However, the Japanese version is output in Japanese,
        # and the English version is output in English.
        # In order to accommodate both, the header is skipped.
        # Instead, columns are indicated by c_sn1/c_sid/c_aid.

l_sn1  = []  # Initialize the list to store Student Names.
l_sid  = []  # Initialize the list to store Student IDs. 
l_ai1  = []  # Initialize the list to store registered attendance IDs (Moodle default)
l_ai2  = []  # Initialize the list to store proper     attendance IDs (Arranged to a form of XX-XX-XX-XX)
l_od1  = []  # Order of students on moodle file.
l_od2  = []  # Order of students who have proper IDs.

i_lp1  =  0  # Number counter for all.
i_lp2  =  0  # Number counter for proper IDs.

for sn1, sid, aid in zip( dt2[c_sn1], dt2[c_sid], dt2[c_aid] ):
            # Loop for students who have registered their attendance ID.

    i_lp1 = i_lp1 + 1

    l_aid = re.split( "[:;]", aid )
            # Separate characters using both ":" and ";".
            # It works even if the record is done in English.

    if ( len(l_aid) >= 15 ):
            # Retrieve the attendance ID only if there is a proper record in Moodle.
        
        ai_tmp = l_aid[ 1] + l_aid[ 3] + "-" + l_aid[ 5] + l_aid[ 7] + "-" \
               + l_aid[ 9] + l_aid[11] + "-" + l_aid[13] + l_aid[15]
            # Reconstruct attendance IDs from Moodle records.

#       ai2    = ai_tmp.replace(" ", "").lower()
        ai2    = ai_tmp.replace(" ", "")
            # - Remove spaces in the string.
            # - lower() changes uppercase letters to lowercase.
            #   Since the letters used for ID are defined in a separate program,
            #   I decided not to lowercase them here.

        i_lp2  = i_lp2 + 1
        i_tmp  = i_lp2

    else:
            # This is the case when a proper attendance ID was not recorded.

        ai2    = "---"
        i_tmp  = -1

    l_sn1.append( sn1   )
    l_sid.append( sid   )
    l_ai1.append( aid   )
    l_ai2.append( ai2   )
    l_od1.append( i_lp1 )
    l_od2.append( i_tmp )
    
    #print( i_lp1, i_tmp,   sn1,            sid,            ai3,            aid )
    #print( i_lp1, i_tmp, l_sn1[i_lp1-1], l_sid[i_lp1-1], l_ai2[i_lp1-1], l_ai1[i_lp1-1] )

  
    ########
    ######## Check duplicate student ID
    ########

if ( len(l_sid) != len(set(l_sid)) ):
        # Check for the same student ID.
        # i.e., Check for two or more records of the same student.
        # "set" is a function that returns non-duplicate elements.
    print( "\n" + "Error: Duplicate student ID numbers exist."             )
    print(        "       When downloading student input from Moodle,"     )
    print(        "       the following option must be selected:"          )
    print(        "       [ Show at most one completed attempt per user ]" )
    print(        "       Debug info.:", len(l_sid), len(set(l_sid)), "\n" )
    sys.exit()

    
############
############ Check duplicate attendance ID
############

l_dup = [1] * len( l_ai2 )
            # Initialize l_dup.
            # =-1: There is a duplication.
            # = 0: No record.
            # = 1: Proper record with no duplication.

for i_nw1 in range(              len(l_ai2) - 1 ):
            # Loop to examine l_ai2 one by one.
            # Not len() but len()-1. 
            # See i_nw2 loop for consistency.
 
    if ( l_dup[i_nw1] <= 0     ):
         continue
            # If l_dup is already 0 or -1, the loop is skipped.

    if ( l_ai2[i_nw1] == "---" ):
         l_dup[i_nw1] =  0
         continue
            # If l_ai2 is not a proper attendance ID, the loop is skipped.
    
    for i_nw2 in range( i_nw1+1, len(l_ai2)     ):
            # Check for duplicate elements in l_ai2 from i_nw1 forward.
            # Not from i_nw1 but from i_nw1+1.
            # See i_nw1 loop for consistency.

        if ( l_ai2[i_nw1] == l_ai2[i_nw2] ):
             l_dup[i_nw1] = -1
             l_dup[i_nw2] = -1
            # If duplicate IDs are found, set both of them to -1.
            # You may want to change from -1 to 0 for easier treatment.
            # Duplicate attendance IDs mean that two or more students
            # share the same attendance ID.

#for i_nw1 in range( len(l_ai2) ):
#    print( i_nw1, l_sid[i_nw1], l_ai2[i_nw1], l_dup[i_nw1], l_sn1[i_nw1] )


############
############ Check matching
############
############ Check that the student's registered attendance ID
############ matches the pre-issued one.
############

l_at1 = [0] * len( l_ai2 )
l_at2 = [0] * len( l_ai2 )
            # Initialize l_at1.  =1: ID found       / =0: ID not found.
            # Initialize l_at2.  =1: Give "present" / =0: Give "absent"

for   i_ai2 in range( len(l_ai2) ):
  for i_ais in range( len(l_ais) ):

    if ( l_ai2[i_ai2] == l_ais[i_ais] ):
         l_at1[i_ai2] =  1
            # If the attendance ID that the student have registered is found,
            # l_at1=1 is given.
         
         if ( l_dup[i_ai2] == 1 ): l_at2[i_ai2] = 1
            # If the student has a unique attedance ID with no duplication,
            # l_at2=1 is given.
         break

 
  #print( i_ai2, l_sid[i_ai2], l_dup[i_ai2], l_at1[i_ai2], l_at2[i_ai2], l_ai2[i_ai2] )


############
############ Sort for output
############

if ( f_in3 == "" ):

    ####
    #### No sorting if the 3rd argument is omitted
    ####

    o_sn1 = l_sn1
    o_sid = l_sid
    o_ai2 = l_ai2
    o_dup = l_dup
    o_at1 = l_at1
    o_at2 = l_at2

else:
    
    ####
    #### Data Loading 3
    #### Student IDs to output
    ####

    dt3    = pandas.read_csv( f_in3, header=None )

    o_sid  = dt3[0].values.tolist()
            # Convert the 1st column of the csv-file to a list.
            # sid: student ID
    
    o_sn1 = [" "] * len( o_sid )
    o_ai2 = [" "] * len( o_sid )
    o_dup = [ 0 ] * len( o_sid )
    o_at1 = [ 0 ] * len( o_sid )
    o_at2 = [ 0 ] * len( o_sid )
            # Initialize for output.

    ####
    #### Sort according to the given student order
    ####

    for   i_otp in range( len(o_sid) ):
      for i_inp in range( len(l_sid) ):
          
        if ( o_sid[i_otp] ==  l_sid[i_inp] ):
               
             o_sn1[i_otp]  =  l_sn1[i_inp]
             o_ai2[i_otp]  =  l_ai2[i_inp]
             o_dup[i_otp]  =  l_dup[i_inp]
             o_at1[i_otp]  =  l_at1[i_inp]
             o_at2[i_otp]  =  l_at2[i_inp]
             break

      #print( i_otp, o_sid[i_otp], o_at2[i_otp] )

    ####
    #### Quick check
    ####

    if ( sum(o_at2) != sum(l_at2) ):
        print( "" )
        print( "Warning: Checksums of o_at2 and l_at2 do not match."                    )
        print( "         Not all student records may be reflected in the final output." )
        print( "" )


############
############ Output
############

f_ot1 = os.path.splitext(f_in2)[0] +     '_at.csv'
            # Output file name.

o_lst = list( zip( o_sid, o_at2, o_ai2, o_dup, o_at1, o_sn1 ) )
            # Bundle multiple lists into one list.

dto   = pandas.DataFrame( o_lst, \
        columns=[ "student_ID", "attendance", "attendance_ID", "dup_check", \
                  "attendance_ID_found", "name" ] )
            # Convert a list into a data frame.
            # index =False to not write out line   names.

dto.to_csv( f_ot1, encoding="utf_8_sig" )
            # Output to csv.
            # Adding "_sig" will output with BOM,
            # so the text will not be garbled when opened in Excel.
